local MODNAME = ...

-------------------------------------------------------------------------------
-- Localized Lua globals
-------------------------------------------------------------------------------
local _G = _G

local string = _G.string
local floor = _G.floor
local concat = _G.table.concat
local format = _G.format

local pairs = _G.pairs
local ipairs = _G.ipairs
local wipe = _G.table.wipe 

-------------------------------------------------------------------------------
-- AddOn namespace
-------------------------------------------------------------------------------
--local MODNAME = "Syllabus"
local Syllabus = LibStub("AceAddon-3.0"):NewAddon(MODNAME, "AceEvent-3.0", "AceConsole-3.0")
_G[MODNAME] = Syllabus

local LDB = LibStub("LibDataBroker-1.1")
local QTip = LibStub("LibQTip-1.0")

local dev = false
--[===[@debug@
dev = true
--@end-debug@]===]
local L = LibStub("AceLocale-3.0"):GetLocale(MODNAME, "enUS", true, dev)

local DataObj
local tooltip

local version = GetAddOnMetadata(MODNAME, "Version")
version = string.gsub(version, "@project.version@", " - Development")

-------------------------------------------------------------------------------
-- Constants
-------------------------------------------------------------------------------
local MY_CLASS, MY_CLASS_UPPER = UnitClass("player")
local MY_LEVEL = UnitLevel("player")
local MY_NAME = UnitName("player")
local MY_RACE = UnitRace("player")
local MY_FACTION = UnitFactionGroup("player")

local COLOR_TABLE = CUSTOM_CLASS_COLORS or RAID_CLASS_COLORS
local CLASS_COLORS = {}

local MY_PROFESSIONS = {}
Syllabus.prof = MY_PROFESSIONS

for k, v in pairs(COLOR_TABLE) do
	CLASS_COLORS[k] = string.format("%2x%2x%2x", v.r * 255, v.g * 255, v.b * 255)
end

-- Profession titles (this will need to be localized
-- The key is the max rank for a skill level, the number is the SpellID where we can get the titla name
-- I'm trying to keep the localization needs to a minimum but I'm not sure it's such a good idea because of
-- the number of GetSpellInfo that are called on the first update
--	[PROFESSIONS_ARCHAEOLOGY]		= { [0] = 80836, [75] = },
local MY_TITLE = {
	[L['Alchemy']]					= { [75] = 2275,  [150] = 2280,  [225] = 3465,  [300] = 11612, [375] = 28597, [450] = 65281, },
	[L['Blacksmithing']]			= { [75] = 2020,  [150] = 2021,  [225] = 3539,  [300] = 9786,  [375] = 29845, [450] = 65282, },
	[L['Enchanting']]				= { [75] = 7414,  [150] = 7415,  [225] = 7416,  [300] = 13921, [375] = 28030, [450] = 65285, },
	[L['Engineering']]			= { [75] = 4039,  [150] = 4040,  [225] = 4041,  [300] = 12657, [375] = 30351, [450] = 61464, },
	[L['Herbalism']]				= { [75] = 2372,  [150] = 2373,  [225] = 3571,  [300] = 11994, [375] = 28696, [450] = 65288, },
	[L['Inscription']]			= { [75] = 45375, [150] = 45376, [225] = 45377, [300] = 45378, [375] = 45379, [450] = 45380, },
	[L['Jewelcrafting']]			= { [75] = 25245, [150] = 24246, [225] = 28896, [300] = 28899, [375] = 28901, [450] = 51310, },
	[L['Leatherworking']]		= { [75] = 2155,  [150] = 2154,  [225] = 3812,  [300] = 10663, [375] = 32550, [450] = 51301, },
	[L['Mining']]					= { [75] = 2581,  [150] = 2582,  [225] = 3568,  [300] = 10249, [375] = 29355, [450] = 50309, },
	[L['Skinning']]				= { [75] = 8615,  [150] = 8619,  [225] = 8620,  [300] = 10769, [375] = 32679, [450] = 50307, },
	[L['Tailoring']]				= { [75] = 3911,  [150] = 3912,  [225] = 3913,  [300] = 12181, [375] = 26791, [450] = 51308, },
	[PROFESSIONS_COOKING]		= { [75] = 2551,  [150] = 3412,  [225] = 54257, [300] = 18261, [375] = 54256, [450] = 65291, },
	[PROFESSIONS_FIRST_AID]		= { [75] = 3279,  [150] = 3280,  [225] = 54254, [300] = 10847, [375] = 54255, [450] = 65292, },
	[PROFESSIONS_FISHING]		= { [75] = 7733,  [150] = 7734,  [225] = 54083, [300] = 18249, [375] = 54084, [450] = 65293, },
}

-------------------------------------------------------------------------------
-- Variables
-------------------------------------------------------------------------------
local update_queued = false
local in_combat = false

local db
local faction_db

-------------------------------------------------------------------------------
-- Functions
-------------------------------------------------------------------------------

-- print functions
function Syllabus:Print(msg) SELECTED_CHAT_FRAME:AddMessage(MODNAME..": "..msg,0.5,1,0) end
function Syllabus:debug(msg,...) if dev then 
		self:err(msg,...)
	end 
end
function Syllabus:err(msg,...)
	geterrorhandler()(format(msg,...) .. " - " .. time())
end

local function print(...) Syllabus:Print(...) end
local function debug(...) Syllabus:debug(...) end
local function err(...) Syllabus:err(...) end


-- Table reuse functions
do
	local table_cache = {}

	-- Returns a table
	function AcquireTable()
		local tbl = tremove(table_cache) or {}
		return tbl
	end

	-- Cleans the table and stores it in the cache
	function ReleaseTable(tbl)
		if not tbl then return end
		
		-- Nested tables ?
		for i=1,#tbl do
			if type(tbl[i]) == "table" then ReleaseTable(tbl[i]) end
		end
		
		wipe(tbl)
		tinsert(table_cache, tbl)
	end
	
	-- Release and Acquire in one go
	function ClearTable(tbl)
		ReleaseTable(tbl)
		return AcquireTable()
	end
end	-- do block

-- Find the base cost from the real cost and the reputation standing
local function GetBaseCost(cost, StandingId)
	assert(StandingId and StandingId > 3,"Invalid reputation StandingId (must be between 4 and 8): "..tostring(StandingId))
	
	local rebate = 1 - (StandingId - 4) * .05
	return floor(cost / rebate)
end

local function GetRealCost(cost, StandingId)
	assert(StandingId and StandingId > 3,"Invalid reputation StandingId (must be between 4 and 8): "..tostring(StandingId))
	
	local rebate = 1 - (StandingId - 4) * .05
	return ceil(cost * rebate)
end

local GetReputationStanding
local GetCheapestCost
do -- Closure for the local tables

	local rep_index_cache = {}
	function GetReputationStanding(rep_name)

		if rep_index_cache[rep_name] then
			local name, _, standingId = GetFactionInfo(rep_index_cache[rep_name])

			if name == rep_name then
				return standingId
			else
				-- Reputations have change, the index is no longer valid
				rep_index_cache[rep_name] = nil
			end
		end

		for factionIndex = 1, GetNumFactions() do
			local name, _, standingId, _, _, _, _, _, isHeader = GetFactionInfo(factionIndex)
			if not isHeader and rep_name == name then
				rep_index_cache[rep_name] = factionIndex
				return standingId
			end
		end

		return 0
	end

	local rep_standings  = {}
	local best_rep = {}
	function GetCheapestCost(cost,rep_table)
		local best_standing = 0

		-- Refesh the rep_standings
		for factionIndex = 1, GetNumFactions() do
			local rep_name, _, standingId, _, _, _, _, _, isHeader = GetFactionInfo(factionIndex)
			if not isHeader then
				-- We refresh the index cache while were's at it
				rep_index_cache[rep_name] = factionIndex
				rep_standings[rep_name] = standingId
			end
		end

		-- Find the best standing
		for rep_name in pairs(rep_table) do
			if rep_standings[rep_name] and rep_standings[rep_name] > best_standing then
				best_standing = rep_standings[rep_name]
			end
		end
		
		if best_standing == 0 then return cost, "" end

		-- Build a string of the best Reputation factions
		wipe(best_rep)
		for rep_name in pairs(rep_table) do
			if rep_standings[rep_name] == best_standing then
				tinsert(best_rep,rep_name)
			end
		end
		sort(best_rep)
		
		return GetRealCost(cost, best_standing), concat(best_rep, ", ")

	end

end -- End of closure

local function ScanProfessionInfo(index)
	if not index then return end
	
	local prof_name, _, rank, max_rank = GetProfessionInfo(index)

	-- Update to current values	
	MY_PROFESSIONS[prof_name] = MY_PROFESSIONS[prof_name] or {}	
	MY_PROFESSIONS[prof_name].rank = rank
	MY_PROFESSIONS[prof_name].max_rank = max_rank
	
	-- We can infer some training from the max rank
	for i=1,#PROFESSION_RANKS do
		local value = PROFESSION_RANKS[i][1]
		if max_rank >= value then 
			-- We know that one
			if type(MY_TITLE[prof_name][value]) ~= 'string' then
				-- Let's keep the calls to GetSpellInfo to a minimum
				local skill_name = GetSpellInfo(MY_TITLE[prof_name][value])
				if skill_name then
					MY_TITLE[prof_name][value] = skill_name
				else
					-- GetSpellInfo didn't return anything
					break
				end
			end
			db.char.known[prof_name..":"..MY_TITLE[prof_name][value]] = true
		else
			break 
		end
	end
	
end

-- This might not need its own function
local function GetUpdatedBrokerText()
	-- Find the professions that need to be trained
	local total_learnable, total_base_cost, total_cheapest_cost = 0, 0, 0
	for my_prof_name, my_prof in pairs(MY_PROFESSIONS) do
		for prof_name, prof in pairs(faction_db.professions) do
			if prof_name == my_prof_name then 
				for skill_name, skill in pairs(prof.skills) do
					--err("pn = %s, sn = %s, mpn = %s",prof_name,skill_name,my_prof_name)
					if not db.char.known[prof_name..':'..skill_name] and 
					   (skill.rank_req == nil or skill.rank_req <= my_prof.rank) and
						(skill.level_req == nil or skill.level_req <= MY_LEVEL) then
						total_learnable = total_learnable + 1
						total_base_cost = total_base_cost + skill.cost
						total_cheapest_cost = total_cheapest_cost + GetCheapestCost(skill.cost,skill.valid_rep)
					end
				end
			end
		end
	end

	-- Find the class skills that need to be trained
	if faction_db.classes[MY_CLASS_UPPER] then
		-- When the database is empty because no trainer have been seen yet
		for skill_name, skill in pairs(faction_db.classes[MY_CLASS_UPPER].skills) do
			if not db.char.known[MY_CLASS_UPPER..':'..skill_name] and 
				(skill.level_req == nil or skill.level_req <= MY_LEVEL) then
				total_learnable = total_learnable + 1
				total_base_cost = total_base_cost + skill.cost
				total_cheapest_cost = total_cheapest_cost + GetCheapestCost(skill.cost,skill.valid_rep)
			end
		end
	end
	
	local text = tonumber(total_learnable)
	if total_cheapest_cost ~= 0 then 
		text = text .. ": " .. GetMoneyString(total_cheapest_cost)
	end

	return text
end

local function UpdateSkills()
	if in_combat then
		update_queued = true
		return
	end
	
	-- Professions related stuff
	wipe(MY_PROFESSIONS)

	local prof1, prof2, arch, fish, cook, firstAid = GetProfessions()

	ScanProfessionInfo(cook)
	ScanProfessionInfo(firstAid)
	ScanProfessionInfo(fish)
	ScanProfessionInfo(arch)
	ScanProfessionInfo(prof1)
	ScanProfessionInfo(prof2)

	-- Class related stuff
	MY_LEVEL = UnitLevel("player")
	
	-- Update the broker text
	DataObj.text = GetUpdatedBrokerText()
end

local yellow_font = CreateFont("LineFont")
yellow_font:CopyFontObject(GameTooltipText)
yellow_font:SetTextColor(1, 0.96, 0.41) -- Yellow

local header_font = CreateFont("HeaderFont")
header_font:CopyFontObject(GameTooltipText)
header_font:SetTextColor(1, 1, 1) -- White

local white_font = CreateFont("WhiteFont")
white_font:CopyFontObject(GameTooltipText)
white_font:SetTextColor(1, 1, 1) -- White

local function ShowTooltip(anchor)
	local line, column

	tooltip = QTip:Acquire(MODNAME.."Tooltip", 3, "LEFT", "CENTER", "RIGHT")
	tooltip:SetAutoHideDelay(0.25, anchor)
	tooltip:SetScale(db.char.tooltip_scale)
	tooltip:SetHeaderFont(header_font)
	tooltip:SetFont(yellow_font)
	tooltip:Clear()
	tooltip:SmartAnchorTo(anchor)
	
	-- The tables to gather the information to display
	local prof_to_display = AcquireTable()
	local class_to_display = AcquireTable()

	local line, column = tooltip:AddHeader()
	tooltip:SetCell(line, 1,
   			string.format("%s - %s %d %s |cff%s%s|r", MY_NAME, LEVEL, MY_LEVEL, MY_RACE, CLASS_COLORS[MY_CLASS_UPPER], MY_CLASS), header_font, "CENTER", 3)
	tooltip:AddSeparator()

	-- Find the professions that need to be trained
	local total_learnable, total_base_cost, total_cheapest_cost = 0, 0, 0
	for my_prof_name, my_prof in pairs(MY_PROFESSIONS) do
		prof_to_display[my_prof_name] = AcquireTable()
		
		for prof_name, prof in pairs(faction_db.professions) do
			if prof_name == my_prof_name then 
				for skill_name, skill in pairs(prof.skills) do
					--err("pn = %s, sn = %s, mpn = %s",prof_name,skill_name,my_prof_name)
					if not db.char.known[prof_name..':'..skill_name] and 
					   (skill.rank_req == nil or skill.rank_req <= my_prof.rank + db.char.skill_range)
					then
					   --err("Adding %s:%s %s",tostring(my_prof_name),tostring(skill_name),tostring(skill.rank_req))
					   local display_skill = AcquireTable()
						display_skill.rank_req = skill.rank_req
						display_skill.level_req = skill.level_req
						display_skill.icon = skill.icon
						display_skill.cost, display_skill.best_rep = GetCheapestCost(skill.cost,skill.valid_rep)
						prof_to_display[my_prof_name][skill_name] = display_skill
						
						-- Running totals
						if (skill.rank_req == nil or skill.rank_req <= my_prof.rank) and
							(skill.level_req == nil or skill.level_req <= MY_LEVEL) then
							total_learnable = total_learnable + 1
							total_base_cost = total_base_cost + skill.cost
							total_cheapest_cost = total_cheapest_cost + display_skill.cost
						end
					end
				end
			end
		end
	end
	
	-- Find the class skills that need to be trained
	if faction_db.classes[MY_CLASS_UPPER] then
		-- When the database is empty because no trainer have been seen yet
		for skill_name, skill in pairs(faction_db.classes[MY_CLASS_UPPER].skills) do
			if not db.char.known[MY_CLASS_UPPER..':'..skill_name] and 
				(skill.level_req == nil or skill.level_req <= MY_LEVEL + db.char.level_range)
			then
				local display_skill = AcquireTable()
				display_skill.rank_req = skill.rank_req
				display_skill.level_req = skill.level_req
				display_skill.icon = skill.icon
				display_skill.cost, display_skill.best_rep = GetCheapestCost(skill.cost,skill.valid_rep)
				class_to_display[skill_name] = display_skill
				
				-- Running totals
				if skill.level_req == nil or skill.level_req <= MY_LEVEL then
					total_learnable = total_learnable + 1
					total_base_cost = total_base_cost + skill.cost
					total_cheapest_cost = total_cheapest_cost + GetCheapestCost(skill.cost,skill.valid_rep)
				end
			end
		end
	end

	-- Display
	-- Sort alpha for the professions
	local sort_prof = AcquireTable()
	for prof_name in pairs(prof_to_display) do
		tinsert(sort_prof,prof_name)
	end
	sort(sort_prof)
	
	local first_line = true
	for _, prof_name in ipairs(sort_prof) do
		local prof = prof_to_display[prof_name]
		--err("prof_name = %s - %s",prof_name, tostring(prof))
		
		if not first_line then
			line, column = tooltip:AddSeparator(4,0,0,0,0)
		else
			first_line = nil
		end
		line, column = tooltip:AddHeader()
		tooltip:SetCell(line, 1, 
			format("%s (%s/%s)",prof_name,MY_PROFESSIONS[prof_name].rank,MY_PROFESSIONS[prof_name].max_rank), 
			header_font,
			"LEFT")
			
		-- Sort profession skills by rank and then alpha
		local sort_skill = AcquireTable()
		for skill_name in pairs(prof) do
			tinsert(sort_skill,skill_name)
			--err("Adding for sort %s:%s",tostring(prof_name),tostring(skill_name))
		end
		sort(sort_skill, function(a,b)
			if prof[a].rank_req ~= prof[b].rank_req then
				return (prof[a].rank_req or 0) < (prof[b].rank_req or 0)
			else
				return a < b
			end
		end)
		
		for _, skill_name in ipairs(sort_skill) do
			local skill = prof[skill_name]
			local text
			local req_text, separator = "", ""
			local can_train = true
			
			--err("Display %s:%s %s",tostring(prof_name),tostring(skill_name),tostring(skill.rank_req))
			if skill.rank_req then
				if skill.rank_req > MY_PROFESSIONS[prof_name].rank then
					req_text = format("%s\124cFFFF0000%s\124r",req_text,skill.rank_req)
					can_train = false
				else
					req_text = format("%s\124cFF00FF00%s\124r",req_text,skill.rank_req)
				end
				separator = ", "
			end
			
			if skill.level_req then
				if skill.level_req > MY_LEVEL then
					req_text = format("%s%s\124cFFFF0000%s %s\124r", req_text, separator, LEVEL, skill.level_req)
					can_train = false
				else
					req_text = format("%s%s\124cFF00FF00%s %s\124r", req_text, separator, LEVEL, skill.level_req)
				end
				--separator = ", "
			end
			
			if req_text ~= "" then req_text = " - "..req_text end
			local line_colour = ""
			if can_train then line_colour = "\124cFF00FF00" end
			--err("Icon: %s",tostring(skill.icon))
			text = format("%s  \124T%s:0:0\124t%s%s",line_colour,skill.icon,skill_name,req_text)

			line, column = tooltip:AddLine()
			tooltip:SetCell(line, 1, text)
			if skill.cost then 
				if db.char.show.reputation then tooltip:SetCell(line, 2, line_colour .. skill.best_rep) end
				tooltip:SetCell(line, 3, GetMoneyString(skill.cost), white_font)
			end
		end
		
		ReleaseTable(sort_skill)
	end
	ReleaseTable(sort_prof)

	-- Only display class section if there is something to display	???
		-- Header for the class
		if not first_line then
			line, column = tooltip:AddSeparator(4,0,0,0,0)
		else
			first_line = nil
		end
		line, column = tooltip:AddHeader()
		tooltip:SetCell(line, 1, format("%s (%s %d)", MY_CLASS, LEVEL, MY_LEVEL), header_font, "LEFT")
	
		-- Sort class skills by level and then alpha
		local sort_class_skill = AcquireTable()
		for skill_name in pairs(class_to_display) do
			tinsert(sort_class_skill,skill_name)
		end
		sort(sort_class_skill, function(a,b)
			if class_to_display[a].level_req ~= class_to_display[b].level_req then
				return (class_to_display[a].level_req or 0) < (class_to_display[b].level_req or 0)
			else
				return a < b
			end
		end)

		for _, skill_name in ipairs(sort_class_skill) do
			local skill = class_to_display[skill_name]
			local text
			local req_text = ""
			local can_train = true

			--err("Display %s:%s %s",tostring(prof_name),tostring(skill_name),tostring(skill.rank_req))
			if skill.level_req then
				if skill.level_req > MY_LEVEL then
					req_text = format("%s\124cFFFF0000%s %s\124r", req_text, LEVEL, skill.level_req)
					can_train = false
				else
					req_text = format("%s\124cFF00FF00%s %s\124r", req_text, LEVEL, skill.level_req)
				end
			end

			if req_text ~= "" then req_text = " - "..req_text end
			--err("Icon: %s",tostring(skill.icon))
			local line_colour = ""
			if can_train then line_colour = "\124cFF00FF00" end
			text = format("%s  \124T%s:0:0\124t%s%s", line_colour, skill.icon, skill_name, req_text)

			line, column = tooltip:AddLine()
			tooltip:SetCell(line, 1, text)
			if skill.cost then 
				if db.char.show.reputation then tooltip:SetCell(line, 2, line_colour .. skill.best_rep) end
				tooltip:SetCell(line, 3, GetMoneyString(skill.cost), white_font)
			end
		end
		ReleaseTable(sort_class_skill)
	
	-- Display totals
	if total_learnable ~= 0 then
		line, column = tooltip:AddLine()
		tooltip:SetCell(line, 1, " ")
		tooltip:AddSeparator()
		line, column = tooltip:AddLine()
		tooltip:SetCell(line, 1, format("\124cFFFF7F00%s\124r",L['Number of skills ready to train:']))
		tooltip:SetCell(line, 3, tonumber(total_learnable), white_font)
	end

	if total_cheapest_cost ~= 0 then
		line, column = tooltip:AddLine()
		tooltip:SetCell(line, 1, format("\124cFFFF7F00%s\124r",L['Total costs to train these skills:']))
		tooltip:SetCell(line, 3, GetMoneyString(total_cheapest_cost), white_font)
	end

	-- Upade LDB text
	DataObj.text = tonumber(total_learnable)
	if total_cheapest_cost ~= 0 then 
		DataObj.text = DataObj.text .. ": " .. GetMoneyString(total_cheapest_cost)
	end
	
	-- Table Clean-up
	ReleaseTable(class_to_display)
	ReleaseTable(prof_to_display)
	
	-- Adjust the hight of the tooltip so that it fits in the screen
	--tooltip:UpdateScrolling(GetScreenHeight() - 30)
	tooltip:UpdateScrolling()
	tooltip:Show()
end

-------------------------------------------------------------------------------
-- Events
-------------------------------------------------------------------------------
function Syllabus:PLAYER_REGEN_DISABLED()
	in_combat = true
end

function Syllabus:PLAYER_REGEN_ENABLED()
	in_combat = false

	if update_queued then
		update_queued = false
		UpdateSkills()
	end
end

function Syllabus:TRADE_SKILL_SHOW()
		-- We don't want to save the information if this isn't our tradeskill list.
	if IsTradeSkillLinked() or IsTradeSkillGuild() then
		return
	end
	DataObj:SetIcon()

	local prof_name = GetTradeSkillLine()
	for i = 1, GetNumTradeSkills() do
		local name, type, _, _, _ = GetTradeSkillInfo(i)

		if name and type ~= "header" then
--			Syllabus:Print("Adding "..name)
			db.char.known[prof_name..':'..name] = true
		end
	end
end

do
	local current_skill = {}	-- Re-usable table for skill insertions.

	local type_filters = {
		["available"] = 0,
		["unavailable"] = 0,
		["used"] = 0,
	}
	local function SavePref()
		-- Save the user's preferences before opening everything.
		for k, v in pairs(type_filters) do
			type_filters[k] = GetTrainerServiceTypeFilter(k) or 0
			SetTrainerServiceTypeFilter(k, 1)
		end

		-- Expend all the headers
		for i = 1, GetNumTrainerServices() do
			local _ , _, category = GetTrainerServiceInfo(i)

			if not category then
				break
			end

			if category == "header" then
				ExpandTrainerSkillLine(i)
			end
		end
	end
	
	local function RestorePref()
		-- Restore the previous values for the user's type filters.
		for k, v in pairs(type_filters) do
			SetTrainerServiceTypeFilter(k, type_filters[k])
		end
	end

	local DataMiner = CreateFrame("GameTooltip", "SyllabusDataMiner", UIParent, "GameTooltipTemplate")
	DataMiner:SetOwner(UIParent, "ANCHOR_NONE")
	
	function Syllabus:TRAINER_SHOW()
		GameTooltip_SetDefaultAnchor(DataMiner, UIParent)
		
		-- Get some information from the trainner tooltip
		DataMiner:SetUnit("npc")

		local trainer_name = _G["SyllabusDataMinerTextLeft1"]:GetText()
		local trainer_line2 = _G["SyllabusDataMinerTextLeft2"]:GetText()
		local trainer_line3 = _G["SyllabusDataMinerTextLeft3"]:GetText()
		-- Some trainers like the DK Class trainers do not have a line 3 i.e. a trainer role
		-- Let's take line 4 for reputation if there is no line 3 
		local trainer_rep = _G["SyllabusDataMinerTextLeft4"]:GetText() or trainer_line3
		DataMiner:Hide()

		local trainer_role = nil
		local trainer_x, trainer_y = GetPlayerMapPosition("player")
		local trainer_zone = GetRealZoneText()
		local trainer_subzone = GetSubZoneText()

		print(string.format("Trainer is: %s <%s>, in %s at %.2f, %.2f. Reputataion = %s", trainer_name, trainer_line2, trainer_zone, trainer_x * 100, trainer_y * 100,trainer_rep))

		SavePref()
		
		-- Get the data from the open Trainer Windows
		local trainer_db
		local is_tradeskill = false
		if IsTradeskillTrainer() then
			trainer_db = faction_db.professions
			is_tradeskill = true
		else
			trainer_db = faction_db.classes
		end

		for idx = 1, GetNumTrainerServices() do
			local svc_name, _, svc_type, svc_texture, svc_reqlevel = GetTrainerServiceInfo(idx)

			if svc_name and svc_type and svc_type ~= "header" then
				local level_req	= GetTrainerServiceLevelReq(idx)
				local skill_req, rank_req, _ = GetTrainerServiceSkillReq(idx)

				if rank_req == 0 then
					skill_req = nil
					rank_req = nil
				end
				local num_reqs = GetTrainerServiceNumAbilityReq(idx)
				local ability_reqs, meets_reqs

				if num_reqs > 0 then
					ability_reqs, meets_reqs = GetTrainerServiceAbilityReq(idx, 1)
				end

				local item_link = GetTrainerServiceItemLink(idx)

				current_skill.item_link = item_link and item_link:match("item:(%d+)") or nil
				current_skill.icon = GetTrainerServiceIcon(idx)
				current_skill.level_req = ((level_req > 0) and level_req) or nil
				--current_skill.description = ((not current_skill.item_link) and GetTrainerServiceDescription(idx)) or nil
				current_skill.cost = GetBaseCost(GetTrainerServiceCost(idx), GetReputationStanding(trainer_rep))
				current_skill.rank_req = rank_req
				current_skill.requires = ability_reqs
				current_skill.svc_header = GetTrainerServiceSkillLine(idx)
				

				local svc_header
				if is_tradeskill then
					svc_header = GetTrainerServiceSkillLine(idx)
				else
					svc_header = MY_CLASS_UPPER
				end

				-- Keep the first svc_name for the trainer
				if not trainer_role then
					trainer_role = svc_header
				elseif trainer_role ~= svc_header then
					err("trainer_role changed from %s to %s",trainer_role,svc_header)
					
				end

				for key, val in pairs(current_skill) do
					trainer_db[svc_header].skills[svc_name][key] = val
				end
				
				trainer_db[svc_header].skills[svc_name].valid_rep[trainer_rep] = true
				trainer_db[svc_header].skills[svc_name].trainers[trainer_name] = true
				
				if svc_type == "used" and svc_name ~= svc_header then
					db.char.known[svc_header..':'..svc_name] = true
				end
			end
		end

		local trainer_info = trainer_db[trainer_role].trainers[trainer_name]
		trainer_info.x = trainer_x
		trainer_info.y = trainer_y
		trainer_info.zone = trainer_zone
		trainer_info.subzone = trainer_subzone
		trainer_info.line2 = trainer_line2 or ""
		trainer_info.line3 = trainer_line3 or ""
		trainer_info.rep = trainer_rep or ""

		RestorePref()
		UpdateSkills()		
	end

	function Syllabus:TRAINER_CLOSED()
		--err("TRAINER_CLOSED called")
		-- On close, we just recheck the values on trained skills
		SavePref()
		
		--local total, total_known = 0, 0
		for idx = 1, GetNumTrainerServices() do
			local header
			local name, subType, category, texture, requiredLevel, topServiceLine = GetTrainerServiceInfo(idx)

			if IsTradeskillTrainer() then
				header = GetTrainerServiceSkillLine(idx)
			else
				header = MY_CLASS_UPPER
			end

			if name and category and category == "used" and name ~= header then
				db.char.known[header..':'..name] = true
				--total_known = total_known + 1
			end
			--total = total + 1
		end
		
		--err("TRAINER_CLOSED: total = %s, total_known = %s", total, total_known)

		RestorePref()
		UpdateSkills()		
	end
end	-- do


-------------------------------------------------------------------------------
-- Initialization functions
-------------------------------------------------------------------------------
function Syllabus:OnInitialize()
	DataObj = LDB:NewDataObject(MODNAME, {
		type	= "data source",
		label	= MODNAME,
		text	= "0",
		icon	= "Interface\\Icons\\INV_Misc_Book_01",
		SetIcon	= function(self, icon)
				  --local num = math.random(15)
				  --self.icon = "Interface\\Icons\\INV_Misc_Book_"..((num < 10) and "0" or "")..num
				  self.icon = icon or "Interface\\Icons\\INV_Misc_Book_01"
			  end,
		OnEnter	= function(display, motion)
				  ShowTooltip(display)
			  end,
		OnLeave	= function(display, motion)
			  end,
	})
	
	-- Defaults describe the data structure
	local defaults = {
		global = {
			Alliance = {
				classes = {
					-- Class Name
					['*'] = {
						trainers = {
							-- Trainer name
							['*'] = {},
						},
						skills = {
							-- Skill name
							['*'] = {
								valid_rep = {},
								trainers = {},
							},
						},
					},
				},
				professions = {
					-- Profession Name
					['*'] = {
						trainers = {
							-- Trainer name
							['*'] = {},
						},
						skills = {
							-- Skill name
							['*'] = {
								valid_rep = {},
								trainers = {},
							},
						},
					},
				},
			},
			Horde = {
				classes = {
					-- Class Name
					['*'] = {
						trainers = {
							-- Trainer name
							['*'] = {},
						},
						skills = {
							-- Skill name
							['*'] = {
								valid_rep = {},
								trainers = {},
							},
						},
					},
				},
				professions = {
					-- Profession Name
					['*'] = {
						trainers = {
							-- Trainer name
							['*'] = {},
						},
						skills = {
							-- Skill name
							['*'] = {
								valid_rep = {},
								trainers = {},
							},
						},
					},
				},
			},
		},
		char = {
			-- Configuration data
			show = {
				-- cost = base or cheapest
				cost = 'cheapest',
				known = false,
				not_trainable = false,
				overall_cost = true,
				requirements = true,
				skill_count = true,
				trainable = true,
				trainable_future = true,
				trainer_names = true,
				reputation = false,
			},
			tooltip_scale = 1,
			level_range = 4,
			skill_range = 10,
			hidden = {},
			known = {
				-- Class or profession name:Skill Name
				['*'] = nil,
			},
		},
	}
	db = LibStub("AceDB-3.0"):New(MODNAME.."DB", defaults)
	Syllabus.db = db
	faction_db = db.global[MY_FACTION]
--	self:SetupOptions()
	self.OnInitialize = nil
end

function Syllabus:OnEnable()
	UpdateSkills()
	self:RegisterEvent("PLAYER_LEVEL_UP", UpdateSkills)
	self:RegisterEvent("PLAYER_REGEN_DISABLED")
	self:RegisterEvent("PLAYER_REGEN_ENABLED")
	self:RegisterEvent("SKILL_LINES_CHANGED", UpdateSkills)
	self:RegisterEvent("TRADE_SKILL_SHOW")
	self:RegisterEvent("TRAINER_SHOW")
	self:RegisterEvent("TRAINER_CLOSED")
end

function Syllabus:OnDisable()
	self:UnregisterEvent("PLAYER_LEVEL_UP")
	self:UnregisterEvent("PLAYER_REGEN_DISABLED")
	self:UnregisterEvent("PLAYER_REGEN_ENABLED")
	self:UnregisterEvent("SKILL_LINES_CHANGED")
	self:UnregisterEvent("TRADE_SKILL_SHOW")
	self:UnregisterEvent("TRAINER_SHOW")
	self:UnregisterEvent("TRAINER_CLOSED")
end
